@@ -14,18 +14,16 @@
  * limitations under the License.
  */
 
+#include "bson/bson.h"
+
 #include <php.h>
-#include <Zend/zend_interfaces.h>
-#include <ext/standard/php_var.h>
 #include <zend_smart_str.h>
+#include <ext/standard/php_var.h>
+#include <Zend/zend_interfaces.h>
 
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
-
-#include "phongo_compat.h"
 #include "php_phongo.h"
-#include "php_bson.h"
+#include "phongo_bson_encode.h"
+#include "phongo_error.h"
 
 zend_class_entry* php_phongo_javascript_ce;
 
@@ -72,6 +70,50 @@ static bool php_phongo_javascript_init_from_hash(php_phongo_javascript_t* intern
 	return false;
 } /* }}} */
 
+HashTable* php_phongo_javascript_get_properties_hash(phongo_compat_object_handler_type* object, bool is_temp) /* {{{ */
+{
+	php_phongo_javascript_t* intern;
+	HashTable*               props;
+
+	intern = Z_OBJ_JAVASCRIPT(PHONGO_COMPAT_GET_OBJ(object));
+
+	PHONGO_GET_PROPERTY_HASH_INIT_PROPS(is_temp, intern, props, 2);
+
+	if (!intern->code) {
+		return props;
+	}
+
+	{
+		zval code;
+
+		ZVAL_STRING(&code, intern->code);
+		zend_hash_str_update(props, "code", sizeof("code") - 1, &code);
+
+		if (intern->scope) {
+			php_phongo_bson_state state;
+
+			PHONGO_BSON_INIT_STATE(state);
+			if (!php_phongo_bson_to_zval_ex(bson_get_data(intern->scope), intern->scope->len, &state)) {
+				zval_ptr_dtor(&state.zchild);
+				goto failure;
+			}
+
+			zend_hash_str_update(props, "scope", sizeof("scope") - 1, &state.zchild);
+		} else {
+			zval scope;
+
+			ZVAL_NULL(&scope);
+			zend_hash_str_update(props, "scope", sizeof("scope") - 1, &scope);
+		}
+	}
+
+	return props;
+
+failure:
+	PHONGO_GET_PROPERTY_HASH_FREE_PROPS(is_temp, props);
+	return NULL;
+} /* }}} */
+
 /* {{{ proto void MongoDB\BSON\Javascript::__construct(string $code[, array|object $scope])
    Construct a new BSON Javascript type. The scope is a document mapping
    identifiers and values, representing the scope in which the code string will
@@ -306,7 +348,30 @@ static PHP_METHOD(Javascript, unserialize)
 	zval_ptr_dtor(&props);
 } /* }}} */
 
+/* {{{ proto array MongoDB\Driver\Javascript::__serialize()
+*/
+static PHP_METHOD(Javascript, __serialize)
+{
+	PHONGO_PARSE_PARAMETERS_NONE();
+
+	RETURN_ARR(php_phongo_javascript_get_properties_hash(PHONGO_COMPAT_OBJ_P(getThis()), true));
+} /* }}} */
+
+/* {{{ proto void MongoDB\Driver\Javascript::__unserialize(array $data)
+*/
+static PHP_METHOD(Javascript, __unserialize)
+{
+	zval* data;
+
+	PHONGO_PARSE_PARAMETERS_START(1, 1)
+	Z_PARAM_ARRAY(data)
+	PHONGO_PARSE_PARAMETERS_END();
+
+	php_phongo_javascript_init_from_hash(Z_JAVASCRIPT_OBJ_P(getThis()), Z_ARRVAL_P(data));
+} /* }}} */
+
 /* {{{ MongoDB\BSON\Javascript function entries */
+/* clang-format off */
 ZEND_BEGIN_ARG_INFO_EX(ai_Javascript___construct, 0, 0, 1)
 	ZEND_ARG_INFO(0, javascript)
 	ZEND_ARG_INFO(0, scope)
@@ -316,6 +381,16 @@ ZEND_BEGIN_ARG_INFO_EX(ai_Javascript___set_state, 0, 0, 1)
 	ZEND_ARG_ARRAY_INFO(0, properties, 0)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_Javascript___toString, 0, 0, IS_STRING, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(ai_Javascript___unserialize, 0, 0, 1)
+	ZEND_ARG_ARRAY_INFO(0, data, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_Javascript_jsonSerialize, 0, 0, IS_ARRAY, 0)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(ai_Javascript_unserialize, 0, 0, 1)
 	ZEND_ARG_INFO(0, serialized)
 ZEND_END_ARG_INFO()
@@ -324,18 +399,19 @@ ZEND_BEGIN_ARG_INFO_EX(ai_Javascript_void, 0, 0, 0)
 ZEND_END_ARG_INFO()
 
 static zend_function_entry php_phongo_javascript_me[] = {
-	/* clang-format off */
 	PHP_ME(Javascript, __construct, ai_Javascript___construct, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(Javascript, __serialize, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Javascript, __set_state, ai_Javascript___set_state, ZEND_ACC_PUBLIC | ZEND_ACC_STATIC)
-	PHP_ME(Javascript, __toString, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Javascript, jsonSerialize, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(Javascript, __toString, ai_Javascript___toString, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(Javascript, __unserialize, ai_Javascript___unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(Javascript, jsonSerialize, ai_Javascript_jsonSerialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Javascript, serialize, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Javascript, unserialize, ai_Javascript_unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Javascript, getCode, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(Javascript, getScope, ai_Javascript_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_FE_END
-	/* clang-format on */
 };
+/* clang-format on */
 /* }}} */
 
 /* {{{ MongoDB\BSON\Javascript object handlers */
@@ -363,9 +439,8 @@ static void php_phongo_javascript_free_object(zend_object* object) /* {{{ */
 
 zend_object* php_phongo_javascript_create_object(zend_class_entry* class_type) /* {{{ */
 {
-	php_phongo_javascript_t* intern = NULL;
+	php_phongo_javascript_t* intern = zend_object_alloc(sizeof(php_phongo_javascript_t), class_type);
 
-	intern = PHONGO_ALLOC_OBJECT_T(php_phongo_javascript_t, class_type);
 	zend_object_std_init(&intern->std, class_type);
 	object_properties_init(&intern->std, class_type);
 
@@ -405,50 +480,6 @@ static int php_phongo_javascript_compare_objects(zval* o1, zval* o2) /* {{{ */
 	return strcmp(intern1->code, intern2->code);
 } /* }}} */
 
-HashTable* php_phongo_javascript_get_properties_hash(phongo_compat_object_handler_type* object, bool is_debug) /* {{{ */
-{
-	php_phongo_javascript_t* intern;
-	HashTable*               props;
-
-	intern = Z_OBJ_JAVASCRIPT(PHONGO_COMPAT_GET_OBJ(object));
-
-	PHONGO_GET_PROPERTY_HASH_INIT_PROPS(is_debug, intern, props, 2);
-
-	if (!intern->code) {
-		return props;
-	}
-
-	{
-		zval code;
-
-		ZVAL_STRING(&code, intern->code);
-		zend_hash_str_update(props, "code", sizeof("code") - 1, &code);
-
-		if (intern->scope) {
-			php_phongo_bson_state state;
-
-			PHONGO_BSON_INIT_STATE(state);
-			if (!php_phongo_bson_to_zval_ex(bson_get_data(intern->scope), intern->scope->len, &state)) {
-				zval_ptr_dtor(&state.zchild);
-				goto failure;
-			}
-
-			zend_hash_str_update(props, "scope", sizeof("scope") - 1, &state.zchild);
-		} else {
-			zval scope;
-
-			ZVAL_NULL(&scope);
-			zend_hash_str_update(props, "scope", sizeof("scope") - 1, &scope);
-		}
-	}
-
-	return props;
-
-failure:
-	PHONGO_GET_PROPERTY_HASH_FREE_PROPS(is_debug, props);
-	return NULL;
-} /* }}} */
-
 static HashTable* php_phongo_javascript_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp) /* {{{ */
 {
 	*is_temp = 1;
@@ -475,6 +506,10 @@ void php_phongo_javascript_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	zend_class_implements(php_phongo_javascript_ce, 1, php_phongo_type_ce);
 	zend_class_implements(php_phongo_javascript_ce, 1, zend_ce_serializable);
 
+#if PHP_VERSION_ID >= 80000
+	zend_class_implements(php_phongo_javascript_ce, 1, zend_ce_stringable);
+#endif
+
 	memcpy(&php_phongo_handler_javascript, phongo_get_std_object_handlers(), sizeof(zend_object_handlers));
 	PHONGO_COMPAT_SET_COMPARE_OBJECTS_HANDLER(javascript);
 	php_phongo_handler_javascript.clone_obj      = php_phongo_javascript_clone_object;
@@ -483,12 +518,3 @@ void php_phongo_javascript_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	php_phongo_handler_javascript.free_obj       = php_phongo_javascript_free_object;
 	php_phongo_handler_javascript.offset         = XtOffsetOf(php_phongo_javascript_t, std);
 } /* }}} */
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: noet sw=4 ts=4 fdm=marker
- * vim<600: noet sw=4 ts=4
- */
