@@ -51,15 +51,18 @@ function drop_collection($uri, $databaseName, $collectionName)
     $command = new Command(['drop' => $collectionName]);
 
     try {
-        /* We need to use WriteConcern::MAJORITY here due to the issue
-         * explained in SERVER-35613: "drop" uses a two phase commit, and due
-         * to that, it is possible that a lock can't be acquired for a
-         * transaction that gets quickly started as the "drop" reaper hasn't
-         * completed yet. */
+        /* Unless we are dropping a collection within the "local" database,
+         * which does not support a write concern, we need to use w:majority due
+         * to the issue explained in SERVER-35613: "drop" uses a two phase
+         * commit, and due to that, it is possible that a lock can't be acquired
+         * for a transaction that gets quickly started as the "drop" reaper
+         * hasn't completed yet. */
+        $wc = $databaseName === 'local' ? new WriteConcern(1) : new WriteConcern(WriteConcern::MAJORITY);
+
         $server->executeCommand(
             $databaseName,
             $command,
-            ['writeConcern' => new WriteConcern(WriteConcern::MAJORITY)]
+            ['writeConcern' => $wc]
         );
     } catch (RuntimeException $e) {
         if ($e->getMessage() !== 'ns not found') {
@@ -95,6 +98,10 @@ function create_test_manager(string $uri = null, array $options = [], array $dri
         $driverOptions['serverApi'] = new ServerApi(getenv('API_VERSION'));
     }
 
+    if (getenv('CRYPT_SHARED_LIB_PATH') && ! isset($driverOptions['autoEncryption']['extraOptions']['cryptSharedLibPath'])) {
+        $driverOptions['autoEncryption']['extraOptions']['cryptSharedLibPath'] = getenv('CRYPT_SHARED_LIB_PATH');
+    }
+
     return new Manager($uri ?? URI, $options, $driverOptions);
 }
 
@@ -219,6 +226,17 @@ function get_uri_option($uri, $option)
     return $matches[1];
 }
 
+/**
+ * Checks that the topology is load balanced.
+ *
+ * @param string $uri
+ * @return boolean
+ */
+function is_load_balanced($uri)
+{
+    return get_primary_server($uri)->getType() === Server::TYPE_LOAD_BALANCER;
+}
+
 /**
  * Checks that the topology is a sharded cluster.
  *
@@ -231,19 +249,19 @@ function is_mongos($uri)
 }
 
 /**
- * Checks that the topology is a sharded cluster using a replica set
+ * Checks that the topology is a sharded cluster using a replica set.
+ *
+ * Note: only the first shard is checked.
  */
-function is_mongos_with_replica_set($uri)
+function is_sharded_cluster_with_replica_set($uri)
 {
-    if (! is_mongos($uri)) {
+    $server = get_primary_server($uri);
+
+    if ($server->getType() !== Server::TYPE_MONGOS && $server->getType() !== Server::TYPE_LOAD_BALANCER) {
         return false;
     }
 
-    $cursor = get_primary_server($uri)->executeQuery(
-        'config.shards',
-        new \MongoDB\Driver\Query([], ['limit' => 1])
-    );
-
+    $cursor = $server->executeQuery('config.shards', new \MongoDB\Driver\Query([], ['limit' => 1]));
     $cursor->setTypeMap(['root' => 'array', 'document' => 'array']);
     $document = current($cursor->toArray());
 
@@ -271,6 +289,8 @@ function is_replica_set($uri)
         return false;
     }
 
+    /* Note: this may return a false negative if replicaSet is specified through
+     * a TXT record for a mongodb+srv connection string. */
     if (get_uri_option($uri, 'replicaSet') === NULL) {
         return false;
     }
@@ -568,84 +588,120 @@ function destroyTemporaryMongoInstance($id = NULL)
     $json = file_get_contents(MONGO_ORCHESTRATION_URI . "/servers/$id", false, $ctx);
 }
 
-function severityToString($type) {
-    switch($type) {
-    case E_DEPRECATED:
-        return "E_DEPRECATED";
-    case E_RECOVERABLE_ERROR:
-        return "E_RECOVERABLE_ERROR";
-    case E_WARNING:
-        return "E_WARNING";
-    case E_NOTICE:
-        return "E_NOTICE";
-    default:
-        return "Some other #_$type";
+/**
+ * Converts an error level (constant or bitmask) to a string description.
+ */
+function severityToString(int $severity): string {
+    static $constants = [
+        'E_ERROR' => E_ERROR,
+        'E_WARNING' => E_WARNING,
+        'E_PARSE' => E_PARSE,
+        'E_NOTICE' => E_NOTICE,
+        'E_CORE_ERROR' => E_CORE_ERROR,
+        'E_CORE_WARNING' => E_CORE_WARNING,
+        'E_COMPILE_ERROR' => E_COMPILE_ERROR,
+        'E_COMPILE_WARNING' => E_COMPILE_WARNING,
+        'E_USER_ERROR' => E_USER_ERROR,
+        'E_USER_WARNING' => E_USER_WARNING,
+        'E_USER_NOTICE' => E_USER_NOTICE,
+        'E_STRICT' => E_STRICT,
+        'E_RECOVERABLE_ERROR' => E_RECOVERABLE_ERROR,
+        'E_DEPRECATED' => E_DEPRECATED,
+        'E_USER_DEPRECATED' => E_USER_DEPRECATED,
+        // E_ALL is handled separately
+    ];
+
+    if ($severity === E_ALL) {
+        return 'E_ALL';
     }
-}
-function raises($function, $type, $infunction = null) {
-    $errhandler = function($severity, $message, $file, $line) {
-        throw new ErrorException($message, 0, $severity, $file, $line);
-    };
 
-    set_error_handler($errhandler, $type);
-    try {
-        $function();
-    } catch(Exception $e) {
-        $exceptionname = get_class($e);
-
-        if ($e instanceof ErrorException && $e->getSeverity() & $type) {
-            if ($infunction) {
-                $trace = $e->getTrace();
-                $function = $trace[0]["function"];
-                if (strcasecmp($function, $infunction) == 0) {
-                    printf("OK: Got %s thrown from %s\n", $exceptionname, $infunction);
-                } else {
-                    printf("ALMOST: Got %s - but was thrown in %s, not %s\n", $exceptionname, $function, $infunction);
-                }
-                restore_error_handler();
-                return $e->getMessage();
-            }
-            printf("OK: Got %s\n", severityToString($e->getSeverity()));
-        } else {
-            printf("ALMOST: Got %s - expected %s\n", get_class($e), $exceptionname);
+    foreach ($constants as $constant => $value) {
+        if ($severity & $value) {
+            $matches[] = $constant;
         }
+    }
+
+    return empty($matches) ? 'UNKNOWN' : implode('|', $matches);
+}
+
+/**
+ * Expects the callable to raise an error matching the expected severity, which
+ * may be a constant or bitmask. May optionally expect the error to be raised
+ * from a particular function. Returns the message from the raised error or
+ * exception, or an empty string if neither was thrown.
+ */
+function raises(callable $callable, int $expectedSeverity, string $expectedFromFunction = null): string
+{
+    set_error_handler(function(int $severity, string $message, string $file, int $line) {
+        throw new ErrorException($message, 0, $severity, $file, $line);
+    });
+
+    try {
+        call_user_func($callable);
+    } catch (ErrorException $e) {
+        if (!($e->getSeverity() & $expectedSeverity)) {
+            printf("ALMOST: Got %s - expected %s\n", severityToString($e->getSeverity()), severityToString($expectedSeverity));
+            return $e->getMessage();
+        }
+
+        if ($expectedFromFunction === null) {
+            printf("OK: Got %s\n", severityToString($e->getSeverity()));
+            return $e->getMessage();
+        }
+
+        $fromFunction = $e->getTrace()[0]['function'];
+
+        if (strcasecmp($fromFunction, $expectedFromFunction) !== 0) {
+            printf("ALMOST: Got %s - but was raised from %s, not %s\n", errorLevelToString($e->getSeverity()), $fromFunction, $expectedFromFunction);
+            return $e->getMessage();
+        }
+
+        printf("OK: Got %s raised from %s\n", severityToString($e->getSeverity()), $fromFunction);
+        return $e->getMessage();
+    } catch (Throwable $e) {
+        printf("ALMOST: Got %s - expected %s\n", get_class($e), ErrorException::class);
+        return $e->getMessage();
+    } finally {
         restore_error_handler();
+    }
+
+    printf("FAILED: Expected %s, but no error raised!\n", ErrorException::class);
+    return '';
+}
+
+/**
+ * Expects the callable to throw an expected exception. May optionally expect
+ * the exception to be thrown from a particular function. Returns the message
+ * from the thrown exception, or an empty string if one was not thrown.
+ */
+function throws(callable $callable, string $expectedException, string $expectedFromFunction = null): string
+{
+    try {
+        call_user_func($callable);
+    } catch (Throwable $e) {
+        if (!($e instanceof $expectedException)) {
+            printf("ALMOST: Got %s - expected %s\n", get_class($e), $expectedException);
+            return $e->getMessage();
+        }
+
+        if ($expectedFromFunction === null) {
+            printf("OK: Got %s\n", $expectedException);
+            return $e->getMessage();
+        }
+
+        $fromFunction = $e->getTrace()[0]['function'];
+
+        if (strcasecmp($fromFunction, $expectedFromFunction) !== 0) {
+            printf("ALMOST: Got %s - but was thrown from %s, not %s\n", $expectedException, $fromFunction, $expectedFromFunction);
+            return $e->getMessage();
+        }
+
+        printf("OK: Got %s thrown from %s\n", $expectedException, $fromFunction);
         return $e->getMessage();
     }
 
-    printf("FAILED: Expected %s thrown!\n", ErrorException::class);
-    restore_error_handler();
-}
-function throws($function, $exceptionname, $infunction = null) {
-    try {
-        $function();
-    } catch (Throwable $e) {
-    } catch (Exception $e) {
-    }
-
-    if (!isset($e)) {
-        echo "FAILED: Expected $exceptionname thrown, but no exception thrown!\n";
-        return;
-    }
-
-    $message = str_replace(array("\n", "\r"), ' ', $e->getMessage());
-    if ($e instanceof $exceptionname) {
-        if ($infunction) {
-            $trace = $e->getTrace();
-            $function = $trace[0]["function"];
-            if (strcasecmp($function, $infunction) == 0) {
-                printf("OK: Got %s thrown from %s\n", $exceptionname, $infunction);
-            } else {
-                printf("ALMOST: Got %s - but was thrown in %s, not %s (%s)\n", $exceptionname, $function, $infunction, $message);
-            }
-            return $e->getMessage();
-        }
-        printf("OK: Got %s\n", $exceptionname);
-    } else {
-        printf("ALMOST: Got %s (%s) - expected %s\n", get_class($e), $message, $exceptionname);
-    }
-
-    return $e->getMessage();
+    printf("FAILED: Expected %s, but no exception thrown!\n", $expectedException);
+    return '';
 }
 
 function printServer(Server $server)
@@ -781,11 +837,6 @@ function failGetMore(Manager $manager)
     $primary = $manager->selectServer(new ReadPreference('primary'));
     $version = get_server_version_from_server($primary);
 
-    if (version_compare($version, "3.2", "<")) {
-        configureFailPoint($manager, 'failReceivedGetmore', 'alwaysOn');
-        return;
-    }
-
     if (version_compare($version, "4.0", ">=")) {
         /* We use 237 here, as that's the same original code that MongoD would
          * throw if a cursor had already gone by the time we call getMore. This
